'=========================================================================
' Prozeduren fr dreidimensionale Graphik
' 3D-GRAPH.BAS
' Uebernommen aus QuickBASIC Januar 2006
' Letzte Ueberarbeitung 29.01.12
'=========================================================================

' Im Hauptprogramm muss das graphische Darstellungsfeld mit den entsprechenden
' mathematischen Fensterkoordinaten initialisiert sein, z.B.:
' SCREEN 18, 32 oder SCREENRES ...
' VIEW (xp1, zp1) - (xp2, zp2)  Pixelkoordinaten der Ecken des Darstellungsfeldes
' WINDOW (x1, z1)-(x2, z2)      mathematische Koordinaten der Ecken des Darstellungsfeldes
' (Die Graphik wird auf die x-z-Ebene projiziert, siehe unten).
' Fuer PlotFktxy und PlotFktrphi sind 32 Bit Farbtiefe erforderlich.


#INCLUDE ONCE "Common.bi"
#INCLUDE ONCE "GRAPHIK.BAS"
#INCLUDE "fbgfx.bi"


DECLARE SUB Projektion(x AS DOUBLE, y AS DOUBLE, z AS DOUBLE, angp AS DOUBLE, _
  BYREF xp AS DOUBLE, BYREF zp AS DOUBLE)
DECLARE SUB Rotation(x AS DOUBLE, y AS DOUBLE, z AS DOUBLE, Achse AS STRING,_
  Winkel AS DOUBLE, BYREF xrot AS DOUBLE, BYREF yrot AS DOUBLE, _
  BYREF zrot AS DOUBLE)
DECLARE SUB Abb3D(x AS DOUBLE, y AS DOUBLE, z AS DOUBLE, gamma AS DOUBLE, _
  angp AS DOUBLE, BYREF xp AS DOUBLE, BYREF zp AS DOUBLE)
DECLARE SUB Koordinaten3D(xmin AS DOUBLE, xmax AS DOUBLE, ymin AS DOUBLE, _
  ymax AS DOUBLE, zmin AS DOUBLE, zmax AS DOUBLE, gamma AS DOUBLE, angp AS DOUBLE, _
  xAchse AS STRING, yAchse AS STRING, zAchse AS STRING)
DECLARE SUB xyzGraphik(x() AS DOUBLE, y() AS DOUBLE, z() AS DOUBLE, gamma AS DOUBLE, _
  angp AS DOUBLE, Farbe AS INTEGER)
DECLARE SUB zrphiGraphik(r() AS DOUBLE, phi() AS DOUBLE, z() AS DOUBLE, _
  gamma AS DOUBLE, angp AS DOUBLE, Farbe AS INTEGER)
DECLARE SUB zrGraphik(r() AS DOUBLE, z() AS DOUBLE, nphi AS INTEGER, angp AS DOUBLE, _
  Farbe AS INTEGER)    
DECLARE SUB PlotFktxy(fkt AS FUNCTION (x As DOUBLE, y As DOUBLE) AS DOUBLE, _
  xmin AS DOUBLE, xmax AS DOUBLE, ymin AS DOUBLE, ymax AS DOUBLE, _
  gamma AS DOUBLE, angp AS DOUBLE)
DECLARE SUB PlotFktrphi(fkt AS FUNCTION (r As DOUBLE, phi As DOUBLE) AS DOUBLE, _
  rmax AS DOUBLE, fktmin AS DOUBLE, fktmax AS DOUBLE, gamma AS DOUBLE, angp AS DOUBLE)
   

SUB Projektion(x AS DOUBLE, y AS DOUBLE, z AS DOUBLE, angp AS DOUBLE, _
  BYREF xp AS DOUBLE, BYREF zp AS DOUBLE)
' Projektion eines Punktes x,y,z unter dem Winkel angp auf den
' Bildschirm (x-z-Ebene, angp in Grad).
' xp, zp sind die Koordinaten des projizierten Punktes.
' Orientierung des Koordinatensystems:
'
'               z ^   y  
'                 I  / 
'                 I / 
'                 I/
'                 +----------> x
    
  DIM AS DOUBLE phi
  
  phi = pi * angp / 180
  xp = x
  zp = z + y * TAN(phi)

END SUB


SUB Rotation(x AS DOUBLE, y AS DOUBLE, z AS DOUBLE, Achse AS STRING,_
  Winkel AS DOUBLE, BYREF xrot AS DOUBLE, BYREF yrot AS DOUBLE, BYREF zrot AS DOUBLE)
' Berechnung der Koordinaten eines Punktes x,y,z nach Rotation
' um eine Koordinatenachse.
' Orientierung des Koordinatensystems wie bei SUB Projektion.
' Parameterliste:
' - x, y, z = Koordinaten vor der Rotation
' - Achse = Achse, um die gedreht werden soll: "x", "y" oder "z"
' - Winkel = Drehwinkel in Grad (entgegen dem Uhrzeigersinn)
' - xrot, yrot, zrot = Koordinaten nach der Rotation

  DIM AS DOUBLE phi
  
  phi = pi * Winkel / 180     'Winkel im Bogenmass
  SELECT CASE LCASE(Achse)
    CASE "x"    'Rotation um x-Achse:
        xrot = x
        yrot = y * COS(phi) - z * SIN(phi)
        zrot = y * SIN(phi) + z * COS(phi)
    CASE "y"    'Rotation um y-Achse:
        xrot = -x * COS(phi) - z * SIN(phi)
        yrot = y
        zrot = -x * SIN(phi) + z * COS(phi)
    CASE "z"    'Rotation um z-Achse:
        xrot = x * COS(phi) - y * SIN(phi)
        yrot = x * SIN(phi) + y * COS(phi)
        zrot = z
  END SELECT

END SUB


SUB Abb3D(x AS DOUBLE, y AS DOUBLE, z AS DOUBLE, gamma AS DOUBLE, _
    angp AS DOUBLE, BYREF xp AS DOUBLE, BYREF zp AS DOUBLE)
' 3D-Abbildung eines Punktes x,y,z ---> xp,zp: Drehung um den Winkel gamma um 
' die z-Achse und nachfolgende Projektion unter dem Winkel angp auf die
' x-z-Ebene

  DIM AS DOUBLE xrot, yrot, zrot
  
  Rotation(x, y, z, "z", gamma, xrot, yrot, zrot)
  Projektion(xrot, yrot, zrot, angp, xp, zp)
  
END SUB


SUB Koordinaten3D(xmin AS DOUBLE, xmax AS DOUBLE, ymin AS DOUBLE, ymax AS DOUBLE, _
  zmin AS DOUBLE, zmax AS DOUBLE, gamma AS DOUBLE, angp AS DOUBLE, xAchse AS STRING, _
  yAchse AS STRING, zAchse AS STRING)
' 3D-Koordinatenfeld wie in der SUB Projektion beschrieben, jedoch zusaetzlich
' Drehung um die z-Achse moeglich.
' Parameter:
' - xmin ... zmax = Wertebereich von x, y und z
' - gamma = Drehwinkel um die z-Achse
' - angp = Projektionswinkel
' - xAchse, yAchse, zAchse = Achsenbeschriftungen
' Die u.a. Graphiken koennen in das Koordinatensystem gezeichnet werden, wenn
' gamma und angp gleich gewaehlt werden. Empfehlung: gamma ca. -30 Grad, angp ca. 30 Grad
  
  DIM AS DOUBLE xpu1, zpu1, xpu2, zpu2, xpu3, zpu3, xpu4, zpu4, _
                xpo1, zpo1, xpo2, zpo2, xpo3, zpo3, xpo4, zpo4, _
                xp01, zp01, xp02, zp02, xp03, zp03, xp04, zp04
    
  'Berechnung der Eckpunkte durch Rotation und Projektion:
  
  'Untere Ebene (z = zmin):
  Abb3D(xmin, ymin, zmin, gamma, angp, xpu1, zpu1)
  Abb3D(xmax, ymin, zmin, gamma, angp, xpu2, zpu2)
  Abb3D(xmax, ymax, zmin, gamma, angp, xpu3, zpu3)
  Abb3D(xmin, ymax, zmin, gamma, angp, xpu4, zpu4)
   
  'Obere Ebene (z = zmax):
  Abb3D(xmin, ymin, zmax, gamma, angp, xpo1, zpo1)
  Abb3D(xmax, ymin, zmax, gamma, angp, xpo2, zpo2)
  Abb3D(xmax, ymax, zmax, gamma, angp, xpo3, zpo3)
  Abb3D(xmin, ymax, zmax, gamma, angp, xpo4, zpo4)
   
  'Eckpunkte der unteren Ebene verbinden:
  LINE (xpu1, zpu1) - (xpu2, zpu2)
  LINE (xpu2, zpu2) - (xpu3, zpu3)
  LINE (xpu3, zpu3) - (xpu4, zpu4)
  LINE (xpu4, zpu4) - (xpu1, zpu1)
   
  'Eckpunkte der oberen Ebene verbinden:
  LINE (xpo3, zpo3) - (xpo4, zpo4)
  LINE (xpo4, zpo4) - (xpo1, zpo1)  
  
  'Untere und obere Eckpunkte verbinden:
  LINE (xpu1, zpu1) - (xpo1, zpo1)
  LINE (xpu3, zpu3) - (xpo3, zpo3)
  LINE (xpu4, zpu4) - (xpo4, zpo4)  
   
  'Beschriftung x-Achse:
  OutText(xpu1, zpu1, STR(CSNG(xmin)), "l", "u")
  OutText(xpu2, zpu2, STR(CSNG(xmax)) + SPACE(1), "l", "u")
  OutText((xpu1 + xpu2) / 2, (zpu1 + zpu2) / 2, xAchse + SPACE(1), "l", "u")
   
  'Beschriftung y-Achse:
  OutText(xpu2, zpu2, SPACE(1) + STR(CSNG(ymin)), "r", "z")
  OutText(xpu3, zpu3, SPACE(1) + STR(CSNG(ymax)), "r", "z")
  OutText((xpu2 + xpu3) / 2, (zpu2 + zpu3) / 2, SPACE(1) + yAchse, "r", "z")
   
  'Beschriftung z-Achse:
  OutText(xpu1, zpu1, STR(CSNG(zmin)) + SPACE(1), "l", "o")
  OutText(xpo1, zpo1, STR(CSNG(zmax)) + SPACE(1), "l", "z")
  OutText((xpu1 + xpo1) / 2, (zpu1 + zpo1) / 2, zAchse + SPACE(1), "l", "z")
  
END SUB


SUB xyzGraphik(x() AS DOUBLE, y() AS DOUBLE, z() AS DOUBLE, gamma AS DOUBLE, _
  angp AS DOUBLE, Farbe AS INTEGER)
' 3D-Graphik einer Funktion z(x, y).
' Parameter:
' - x von x(0) bis x(m)
' - y von y(0) bis y(n)
' - z von z(0,0) bis z(m, n)
' - gamma = Drehwinkel um die z-Achse
' - angp = Projektionswinkel
' - Farbe = Farbe der gezeichneten Linien
      
  DIM AS INTEGER m = UBOUND(x), n = UBOUND(y), i, j
  DIM AS DOUBLE xp(m, n), zp(m, n)
    
  'Berechnung der Funktionspunkte durch Rotation und Projektion:
  FOR i = 0 TO m
    FOR j = 0 TO n
      Abb3D(x(i), y(j), z(i, j), gamma, angp, xp(i, j), zp(i, j))
    NEXT j
  NEXT i
  
  'Oberflaechenlinien der Funktion fuer x = const:
  FOR i = 0 TO m
    FOR j = 1 TO n
      LINE (xp(i, j-1), zp(i, j-1)) - (xp(i, j), zp(i, j)), Farbe
    NEXT j
  NEXT i
    
  'Oberflaechenlinien der Funktion fuer y = const:
  FOR j = 0 TO n
    FOR i = 1 TO m
      LINE (xp(i-1, j), zp(i-1, j)) - (xp(i, j), zp(i, j)), Farbe
    NEXT i
  NEXT j
  
END SUB


SUB zrphiGraphik(r() AS DOUBLE, phi() AS DOUBLE, z() AS DOUBLE, _
                 gamma AS DOUBLE, angp AS DOUBLE, Farbe AS INTEGER)
' Funktion z(r, phi) in einem x-y-z-Koordinatensystem.
' Parameter:
' - Variable r von r(0, 0) bis r(m, n)
' - Variable phi von phi(0) bis phi(m) im Bogenmass
' - Variable z von z(0) bis z(n)
' - Farbe = Farbe der gezeichneten Linien
' - gamma = Drehwinkel um die z-Achse
' - angp = Projektionswinkel (siehe SUB Projektion).
' Die Funktion wird fuer konstante "Hoehenlinien" z(0) bis z(n) gezeichnet. 
' Die SUB eignet sich fuer Funktionen der Form z(r, phi) = f(r) * h(phi).
' Dann laesst sich im Hauptprogramm fuer jeden Wert z(j) und einen Wert phi(i)
' ein Wert r(i) = g(z(j)) / h(phi(i)) berechnen, wobei g die Umkehrfunktion
' von f ist.

  DIM AS INTEGER m  = UBOUND(phi), n = UBOUND(z), i, j
  DIM AS DOUBLE x(m, n), y(m, n), zneu(m, n), xp(m, n), zp(m, n)

  'Berechnung von z(r,phi) in kartesischen Koordinaten:

  FOR i = 0 TO m
    FOR j = 0 TO n
      x(i, j) = r(i, j) * COS(phi(i))
      y(i, j) = r(i, j) * SIN(phi(i))
      zneu(i, j) = z(j)
    NEXT j
  NEXT i

  'Rotation und Projektion:
  FOR i = 0 TO m
    FOR j = 0 TO n
      Abb3D(x(i, j), y(i, j), zneu(i, j), gamma, angp, xp(i, j), zp(i, j))
    NEXT j
  NEXT i

  'Linien fuer z = const:
  FOR j = 0 TO n
    FOR i = 1 TO m
      LINE (xp(i-1, j), zp(i-1, j)) - (xp(i, j), zp(i, j)), Farbe
    NEXT i
  NEXT j
  
  'Linien fuer phi = const:
  FOR i = 0 TO m
    FOR j = 1 TO n
      LINE (xp(i, j-1), zp(i, j-1)) - (xp(i, j), zp(i, j)), Farbe
    NEXT j
  NEXT i

END SUB


SUB zrGraphik(r() AS DOUBLE, z() AS DOUBLE, m AS INTEGER, angp AS DOUBLE, _
              Farbe AS INTEGER)
' Funktion z(r) in einem x-y-z-Koordinatensystem.
' Parameter:
' - Variable r von r(0) bis r(n)
' - Variable z von z(0) bis z(n)
' - m = oberer Feldindex fuer phi() (wird in der SUB intern erzeugt)  
' - Farbe = Farbe der gezeichneten Linien
' - angp = Projektionswinkel (siehe SUB Projektion).
' Die Funktion wird fuer konstante Radien r(0) bis r(n) gezeichnet.
' Die SUB eignet sich fuer rotationssymmetrische Funktionen.
  
  DIM AS INTEGER n = UBOUND(r), i, j
  DIM AS DOUBLE phi(m), x(m, n), y(m, n), zneu(m, n), xp(m, n), zp(m, n)
                
  'Berechnung von z(r,phi) in kartesischen Koordinaten:
  FOR i = 0 TO m
    phi(i) = 2 * pi * i / m
    FOR j = 0 TO n
        x(i, j) = r(j) * COS(phi(i))
        y(i, j) = r(j) * SIN(phi(i))
        zneu(i, j) = z(j)
    NEXT j
  NEXT i

  'Projektion:
  FOR i = 0 TO m
    FOR j = 0 TO n
      Projektion(x(i, j), y(i, j), zneu(i, j), angp, xp(i, j), zp(i, j))
    NEXT j
  NEXT i

  'Linien fuer r = const:
  FOR i = 0 TO m
    FOR j = 1 TO n
      LINE (xp(i, j-1), zp(i, j-1)) - (xp(i, j), zp(i, j)), Farbe
    NEXT j
  NEXT i
    
  'Linien fuer phi = const:
  FOR i = 1 TO m
    FOR j = 0 TO n
      LINE (xp(i-1, j), zp(i-1, j)) - (xp(i, j), zp(i, j)), Farbe
    NEXT j
  NEXT i

END SUB


SUB PlotFktxy(fkt AS FUNCTION (x As DOUBLE, y As DOUBLE) AS DOUBLE, _
  xmin AS DOUBLE, xmax AS DOUBLE, ymin AS DOUBLE, ymax AS DOUBLE, _
  gamma AS DOUBLE, angp AS DOUBLE)
' 3D-Plot einer Funktion fkt(x, y).
' Parameter:
' - xmin, xmax, ymin, ymax = Wertebereich von x und y
' - gamma = Drehwinkel um die z-Achse (ca. -20 bis -30 Grad)
' - angp = Projektionswinkel (ca. 20 bis 30 Grad)
' Benoetigte Prozedur: FUNCTION fkt (x, y) (im Hauptmodul zu definieren)
' Aufruf der SUB im Hauptmodul: PlotFkt(@fkt, (sonstige Parameter))
' Um den Plot kann mit der SUB Koordinaten3D ein Koordinatensystem
' gezeichnet werden.
  
  DIM AS DOUBLE x, y, dx, dy, z, zmin, zmax, xp, zp
  DIM AS INTEGER x1, y1, x2, y2
  
  'Berechnen von dx und dy, so dass diese mit Sicherheit < Pixelgroesse:
  USING FB
  SCREENCONTROL GET_VIEWPORT, x1, y1, x2, y2
  dx = 0.5 * (xmax - xmin) / (x2 - x1)
  dy = 0.5 * (ymax - ymin) / (y2 - y1)
    
  'Groessten und kleinsten Wert der Funktion suchen:
  zmin = 0
  zmax = 0
  FOR x = xmin TO xmax STEP dx
    FOR y = ymin TO ymax STEP dy
      z = fkt(x, y)
      IF z > zmax THEN zmax = z
      IF z < zmin THEN zmin = z
    NEXT y
  NEXT x
  
  'Berechnung der Funktionspunkte durch Rotation und Projektion und zeichnen.
  'Die Zeichnung erfolgt von hinten nach vorn!
  FOR y = ymax TO ymin STEP -dy
    FOR x = xmin TO xmax STEP dx
      z = fkt(x, y)
      Abb3D(x, y, z, gamma, angp, xp, zp)
      PSET (xp, zp), RGBWert(0.5 * CINT((z - zmin) * 20 / (zmax - zmin)))
    NEXT x
  NEXT y
  
END SUB


SUB PlotFktrphi(fkt AS FUNCTION (r As DOUBLE, phi As DOUBLE) AS DOUBLE, _
  rmax AS DOUBLE, fktmin AS DOUBLE, fktmax AS DOUBLE, gamma AS DOUBLE, angp AS DOUBLE)
' 3D-Plot einer Funktion fkt(r, phi).
' Parameter:
' - rmax: Wertebereich von r = 0 bis rmax (phi von 0 bis 2 * pi)
' - fktmin, fktmax: minimaler und maximaler Wert fuer die Abbildung (Werte darunter
'   und darueber werden nicht gezeichnet)
' - gamma = Drehwinkel um die z-Achse (ca. -20 bis -30 Grad)
' - angp = Projektionswinkel (ca. 20 bis 30 Grad)
' Benoetigte Prozedur: FUNCTION fkt (r, phi) (im Hauptmodul zu definieren)
' Aufruf der SUB im Hauptmodul: PlotFkt(@fkt, (sonstige Parameter))
' Um den Plot kann mit der SUB Koordinaten3D ein Koordinatensystem
' gezeichnet werden.
  
  DIM AS DOUBLE r, phi, dr, dphi, phi1, phi2, z, zmin, zmax, xp, zp
  DIM AS INTEGER x1, y1, x2, y2
  
  'Berechnen von dr und dphi, so dass mit Sicherheit Schrittweite < Pixelgroesse:
  USING FB
  SCREENCONTROL GET_VIEWPORT, x1, y1, x2, y2 
  dr = 0.5 * rmax / (x2 - x1)
  dphi = 0.5 * pi / (y2 - y1)
    
  'Groessten und kleinsten Wert der Funktion suchen:
  zmin = 0
  zmax = 0
  FOR r = 0 TO rmax STEP dr
    FOR phi = 0 TO 2 * pi STEP dphi
      z = fkt(r, phi)
      IF z > zmax THEN
        IF z > fktmax THEN zmax = fktmax ELSE zmax = z
      END IF
      IF z < zmin THEN
        IF z < fktmin THEN zmin = fktmin ELSE zmin = z
      END IF
    NEXT phi
  NEXT r
  
  'Berechnung der Funktionspunkte durch Rotation und Projektion und zeichnen.
  'Die Zeichnung erfolgt von hinten nach vorn!
  phi1 = -gamma * pi / 180
  phi2 = phi1 + pi
  FOR r = rmax TO 0 STEP -dr
    FOR phi = phi1 TO phi2 STEP dphi
      z = fkt(r, phi)
      IF z <= fktmax AND z >= fktmin THEN
        Abb3D(r * COS(phi), r * SIN(phi), z, gamma, angp, xp, zp)
        PSET (xp, zp), RGBWert(0.5 * CINT((z - zmin) * 20 / (zmax - zmin)))
      END IF
    NEXT phi
  NEXT r
  phi1 = phi1 + pi
  phi2 = phi1 + pi
  FOR r = 0 TO rmax STEP dr
    FOR phi = phi1 TO phi2 STEP dphi
      z = fkt(r, phi)
      IF z <= fktmax AND z >= fktmin THEN
        Abb3D(r * COS(phi), r * SIN(phi), z, gamma, angp, xp, zp)
        PSET (xp, zp), RGBWert(0.5 * CINT((z - zmin) * 20 / (zmax - zmin)))
      END IF
    NEXT phi
  NEXT r  
  
END SUB
